# coding: utf-8

from __future__ import division
import numpy as np
from sklearn import metrics
from . import utils
import pandas


def calculate_measures(results, measures):
    processed_result = {}
    supports = []
    for (measure_name, measure_function) in measures:

        measure_result, pvalue, support = apply_measure(results, measure_function)
        processed_result[measure_name] = measure_result
        processed_result[measure_name + "_p_val"] = pvalue
        supports.append(support)

    assert(utils.all_equal(supports))

    support = supports[0]
    processed_result.update(dict(support=support))

    return processed_result


def apply_measure(results, measure_function):
    measure_result, support = calculate_result_for_measure(results, measure_function)
    if "permutations" in results and len(results["permutations"].keys()) > 0:
        permutation_values = [calculate_result_for_measure(perm_res, measure_function)[0] for perm_id, perm_res in results["permutations"].iteritems()]
        pvalue = len([p for p in permutation_values if p >= measure_result]) * 1.0 / (len(permutation_values))
    else:
        pvalue = np.nan

    return measure_result, pvalue, support


def auc_fn(actual, predicted_probabilities, categories):
    return metrics.roc_auc_score(actual, predicted_probabilities)


def accuracy_fn(actual, predicted_probabilities, categories):
    predicted = [1 if p > 0.5 else 0 for p in predicted_probabilities]
    return metrics.accuracy_score(actual, predicted)


def y_indices_from_fold_result(fold_results):
    return np.array(utils.flatten([fold_res["y_ids"] for k, fold_res in fold_results.iteritems()]))


def y_true_from_fold_result(fold_results):
    return np.array(utils.flatten([fold_res["actual"] for k, fold_res in fold_results.iteritems()]))


def y_score_from_fold_result(fold_results):
    return np.array(utils.flatten([fold_res["predicted_probabilities"][:, 1] for k, fold_res in fold_results.iteritems()]))


def calculate_result_for_measure(results, measure):
    categories = results["categories"]
    fold_results = results["fold_results"]

    actual = y_true_from_fold_result(fold_results)
    predicted_probabilities = y_score_from_fold_result(fold_results)

    measure_result = measure(actual, predicted_probabilities, categories)
    support = [sum(actual == c) for c in categories]

    return measure_result, list(zip(categories, support))


def feature_importances(fold_results):
    all_feature_importances = []

    for fold, fold_result in fold_results.iteritems():
        all_feature_importances.append(fold_result["classifier_weights"])

    feature_importances_by_folds = np.array(all_feature_importances)
    return feature_importances_by_folds.mean(axis=0), feature_importances_by_folds.std(axis=0)


def feature_importances_table(fold_results, feature_mapping, session_id, subject):
    importance_means, importance_stds = feature_importances(fold_results)
    importances_res = []

    for feature_id, (importance_mean, importance_std) in enumerate(zip(importance_means, importance_stds)):
        info = feature_mapping[feature_id].copy()
        info["feature_importances_folds_mean"] = importance_mean
        info["feature_importances_folds_std"] = importance_std
        info["session"] = session_id
        info["subject"] = subject
        importances_res.append(info)

    features_table = pandas.DataFrame(importances_res)
    return features_table



# coding=utf-8

from . import interval
from . import system


class WaveSurfer():
    def __init__(self, intervals):
        self.intervals = intervals
        self.counter = 0

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and self.intervals == other.intervals)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        for i in self.intervals:
            yield i

    def diff(self, x):
        differences = []
        if len(self.intervals) != len(x.intervals):
            system.warning("number of intervals: {} vs {}".format(len(self.intervals), len(x.intervals)))

        for i in x:
            if i.start != self.current().start:
                system.error("different starts on interval ({}) {} vs {}".format(self.counter + 1, i.start, self.current().start))
                raise Exception("Intervals are not the same")

            if i.value != self.current().value:
                differences.append((self.counter + 1, i.start, i.end))
            next(self)

        return differences

    def current(self):
        return self.intervals[self.counter]

    def next(self, value=None):
        if self.has_next():
            self.counter += 1
            val = self.intervals[self.counter]
            if value and value != val.value:
                return self.next(value)
            else:
                return val
        else:
            return None

    def values_in_range(self, start, end):
        values = []
        for i in self.intervals:
            if i.end < start:
                continue
            elif i.start > end:
                break
            else:
                values.append(i.value)
        return values

    def next_until(self, time):
        if self.current().end < time and self.has_next():
            next(self)
            return self.next_until(time)
        else:
            return self.current()

    def has_next(self):
        return self.counter < len(self.intervals) - 1

    def __repr__(self):
        return "Wavesurfer: " + str(self.intervals)

    def as_vad(self):
        vad_intervals = []
        current_ipu = None
        for i in self:
            if i.value == "#":
                if current_ipu:
                    vad_intervals.append(current_ipu)
                    current_ipu = None
                vad_intervals.append(interval.Interval(i.start, i.end, "0"))

            else:
                if not current_ipu:
                    current_ipu = interval.Interval(i.start, i.end, "1")
                else:
                    current_ipu = interval.Interval(current_ipu.start, i.end, "1")

        return WaveSurfer(vad_intervals)



# coding: utf-8

import mne
import numpy as np
from . import system


def from_eeglab_raw_setfile(set_filename):
    data_mne = mne.io.read_raw_eeglab(set_filename)
    data_mne.load_data()
    d = data_mne._data * 1e6
    info = data_mne.info

    system.info("Loading data from {} (channels={}, samples={})".format(set_filename, d.shape[0], d.shape[1]))
    system.info(str(info))
    return info, d


def from_eeglab_epochs_setfile(set_filename):
    data_mne = mne.io.read_epochs_eeglab(set_filename, events=None, event_id=None)
    d = data_mne.get_data() * 1e6
    info = data_mne.info

    system.info("Loading data from {} (epochs={}, channels={}, samples={})".format(set_filename, d.shape[0], d.shape[1], d.shape[2]))
    system.info(str(info))

    return info, d


def read_montage(montage_filename):
    montage = mne.channels.read_montage(montage_filename)
    new_locs = np.array([(-y, x, z) for (x, y, z) in montage.pos])
    montage.pos = new_locs
    return montage



# coding=utf-8

from . import signal_processing
from . import feature_extraction
from . import utils
from . import data_import

import numpy as np
from scipy import signal
import mne


class EEGFeatureExtractor(feature_extraction.FeatureExtractor):
    def __init__(self, config):
        self.params = utils.read_config(config)
        self.time_limits = self.params["time_limits"]
        self.tmin = self.params["tmin"]
        self.freq = float(self.params["freq"])
        self.n_channels = int(self.params["n_channels"])
        self.channels_to_extract_from = list(self.params["channels_to_extract_from"])
        self.verbose = self.params["verbose"] if ("verbose" in self.params) else None
        if self.verbose:
            print(self.params)


class RawExtractor(EEGFeatureExtractor):
    def __init__(self, config):
        super(RawExtractor, self).__init__(config)

    def extract(self, trial):
        assert self.n_channels == trial.shape[0], "the number of channels in the extractor configuration doen't match {} vs {}".format(self.n_channels, trial.shape[0])
        ti, tf = self.time_limits

        initial_frame = int((ti - self.tmin) * self.freq)
        end_frame = int((tf - self.tmin) * self.freq)
        data = trial[self.channels_to_extract_from, initial_frame:end_frame]
        return data

    def feature_names(self):
        pass


class TrashExtractor(EEGFeatureExtractor):
    def __init__(self, config):
        super(TrashExtractor, self).__init__(config)

    def extract(self, trial):
        assert self.n_channels == trial.shape[0], "the number of channels in the extractor configuration doen't match {} vs {}".format(self.n_channels, trial.shape[0])
        ti, tf = self.time_limits

        initial_frame = int((ti - self.tmin) * self.freq)
        end_frame = int((tf - self.tmin) * self.freq)
        data = trial[self.channels_to_extract_from, initial_frame:end_frame]

        data = np.random.random(data.shape[0])
        return data

    def feature_names(self):
        pass


class FreqExtractor(EEGFeatureExtractor):
    def __init__(self, config):
        super(FreqExtractor, self).__init__(config)
        self.min_freq = self.params["min_freq"]
        self.max_freq = self.params["max_freq"]

    def extract(self, trial):
        assert self.n_channels == trial.shape[0], "the number of channels in the extractor configuration doen't match {} vs {}".format(self.n_channels, trial.shape[0])
        ti, tf = self.time_limits

        initial_frame = int((ti - self.tmin) * self.freq)
        end_frame = int((tf - self.tmin) * self.freq)
        data = trial[self.channels_to_extract_from, initial_frame:end_frame]

        freqs_by_channel = []

        for ch in self.channels_to_extract_from:
            x = data[ch, :]
            freqs, Pxx = signal.welch(x, fs=self.freq)
            idx_30hz = np.argmax(freqs > self.max_freq)  # No more than x hz
            idx_1hz = np.argmin(freqs < self.min_freq)  # No less than y Hz

            freqs = freqs[idx_1hz:idx_30hz]
            presence_of_freq = Pxx[idx_1hz:idx_30hz]

            freqs_by_channel.extend(presence_of_freq)
        res = np.array(freqs_by_channel)
        return res

    def mapping(self):
        trial = np.zeros((self.n_channels, 1075))  # (channels x samples)
        montage = data_import.read_montage('/home/pbrusco/projects/montages/LNI.sfp')

        ti, tf = self.time_limits

        initial_frame = int((ti - self.tmin) * self.freq)
        end_frame = int((tf - self.tmin) * self.freq)
        data = trial[self.channels_to_extract_from, initial_frame:end_frame]
        res = {}

        i = 0
        for ch in self.channels_to_extract_from:
            x = data[ch, :]
            freqs, Pxx = signal.welch(x, fs=self.freq)
            idx_30hz = np.argmax(freqs > self.max_freq)  # No more than x hz
            idx_1hz = np.argmin(freqs < self.min_freq)  # No less than y Hz

            freqs = freqs[idx_1hz:idx_30hz]
            for freq in freqs:
                res[i] = dict(channel=ch,
                              channel_name=montage.ch_names[ch],
                              freq=freq,
                              channel_position=montage.pos[ch],
                              name="c{}|{}Hz)".format(ch, round(freq, 2))
                              )
                i = i + 1

        return res

    def feature_names(self):
        return [c["name"] for c in self.mapping().values()]


class WaveletsExtractor(EEGFeatureExtractor):
    def __init__(self, config):
        super(WaveletsExtractor, self).__init__(config)

    def extract(self, trial):
        pass


class WindowedExtractor(EEGFeatureExtractor):
    def __init__(self, config):
        super(WindowedExtractor, self).__init__(config)
        self.window_sizes = self.params["window_sizes"]
        self.step_in_frames = self.params["step_in_frames"]

    def extract(self, trial):
        # Trial shape: (channels x samples)
        assert self.n_channels == trial.shape[0], "the number of channels in the extractor configuration doen't match {} vs {}".format(self.n_channels, trial.shape[0])
        features = []
        for window_size_in_secs in self.window_sizes:
            ws_frames = int(window_size_in_secs * self.freq)
            if self.verbose:
                print("ws frames: ", ws_frames)
            for starting_sample, chunk in signal_processing.sliding_window(trial,
                                                                           window_size_in_frames=ws_frames,
                                                                           step_in_frames=self.step_in_frames,
                                                                           time_limits=self.time_limits,
                                                                           t0=self.tmin,
                                                                           freq=self.freq,
                                                                           debugging=self.verbose,
                                                                           ):
                window_coefficients = []
                for ch in self.channels_to_extract_from:
                    y = chunk[ch, :]
                    window_coefficients.append(y.mean())
                features.extend(window_coefficients)

        features = np.array(features)
        return features

    def mapping(self):
        i = 0
        dummy_epoch = np.zeros((self.n_channels, 1075))  # (channels x samples)
        montage = data_import.read_montage(self.params["montage"])

        res = {}
        for window_size_in_secs in self.window_sizes:
            ws_frames = int(window_size_in_secs * self.freq)
            for starting_sample, chunk in signal_processing.sliding_window(dummy_epoch,
                                                                           window_size_in_frames=ws_frames,
                                                                           step_in_frames=self.step_in_frames,
                                                                           time_limits=self.time_limits,
                                                                           t0=self.tmin,
                                                                           freq=self.freq,
                                                                           ):
                for ch in self.channels_to_extract_from:
                    t0, t1 = signal_processing.frame_time_limits(starting_sample, self.freq, self.tmin, ws_frames)
                    res[i] = dict(
                        channel=ch,
                        channel_name=montage.ch_names[ch],
                        starting_sample=starting_sample,
                        window_size=window_size_in_secs,
                        starting_time=round(t0, 3),
                        end_time=round(t1, 3),
                        channel_position=montage.pos[ch],
                        name="(w{}|c{}|{}|{})".format(round(window_size_in_secs, 2), ch, round(t0, 2), round(t1, 2))
                    )
                    i = i + 1
        return res

    def feature_names(self):
        return [c["name"] for c in self.mapping().values()]



# coding: utf-8

import sklearn.cross_validation
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from . import utils


class ClassificationHelper(object):
    def __init__(self, config, classes=[0, 1], classifier=None):
        self.config = config

        if classifier:
            self.classifier = classifier
        else:
            self.classifier = self._build_classifier()
            print("Using {} classifier".format(self.classifier))

        self.classes = classes

    def _build_classifier(self):
        classifiers = {"SVM": SVC(kernel="linear", probability=True, C=1, class_weight='balanced'),  # Balanced da pesos según cantidad de instancias
                       "SVMG": SVC(kernel='rbf', probability=True, gamma=0.7, C=1, class_weight='balanced'),
                       "RF": RandomForestClassifier(n_estimators=self.config["classifier_n_trees"], criterion='gini', n_jobs=-1, random_state=self.config["seed"], class_weight='balanced'),
                       "LDA": LinearDiscriminantAnalysis(),
                       }

        return classifiers[self.config["classifier_name"]]

    def classification_probas(self, X_train, y_train, X_test):
        self.classifier.fit(X_train, y_train)
        return self.classifier.predict_proba(X_test)

    def one_speaker_out_cross_validation(self, X, y, speakers, verbose=True):
        folds = sklearn.cross_validation.LeaveOneLabelOut(speakers)
        results = {}
        results = {}

        for i, (train_index, test_index) in enumerate(folds):
            if verbose:
                utils.print_inline("speaker: {} of {} for {}".format(i + 1, len(set(speakers)), self.config["classifier_name"]))
            X_train, X_test = X[train_index], X[test_index]
            y_train, y_test = y[train_index], y[test_index]

            results[i] = self._classify(X_train, X_test, y_train)
            results[i]["actual"] = y_test
            results[i]["y_ids"] = test_index

        return {"categories": self.classes, "results": results, "y": y}

    def k_fold_cross_validation(self, X, y, n_folds, verbose=True, subsample=False):
        counts = [sum(y == c) for c in self.classes]

        if verbose:
            print("running classifiers for", list(zip(self.classes, counts)))
        folds = sklearn.cross_validation.StratifiedKFold(y, n_folds=n_folds, random_state=self.config["seed"])

        fold_results = {}

        for i, (train_index, test_index) in enumerate(folds):
            if verbose:
                utils.print_inline("K-fold: {} of {} for {}".format(i + 1, n_folds, self.config["classifier_name"]))
            X_train, X_test = X[train_index], X[test_index]
            y_train, y_test = y[train_index], y[test_index]

            if subsample:
                sub_ids = utils.subsample_ids(y_train)
                X_train = X_train[sub_ids]
                y_train = y_train[sub_ids]

            fold_results[i] = self._classify(X_train, X_test, y_train)
            fold_results[i]["actual"] = y_test
            fold_results[i]["y_ids"] = test_index

        return {"categories": self.classes, "fold_results": fold_results, "y": y}

    def _classify(self, X_train, X_test, y_train):
        self.classifier.fit(X_train, y_train)
        y_pred_probabilities = self.classifier.predict_proba(X_test)
        res = {"predicted_probabilities": y_pred_probabilities}

        if self.config["save_weights"]:
            if self.config["classifier_name"] == "RF":
                feature_importances = self.classifier.feature_importances_
            elif "SVM" in self.config["classifier_name"]:
                feature_importances = self.classifier.coef_
            else:
                raise "I don't know whitch weights to save"
            res.update({"classifier_weights": feature_importances})
        return res
